@@ -133,8 +133,9 @@
                  (init (vector-ref v 3))
                  (end (vector-ref v 4)))
              (if init (init cfg))
-             (apply proc cfg spec args)
-             (if end (end cfg)))))
+             (let ((res (apply proc cfg spec args)))
+               (if end (end cfg))
+               res))))
      ((null? (cdr args))
       (app-help spec args)
       (error "Expected a command"))
@@ -335,7 +336,7 @@
   (define (next-prefix prefix name)
     (append (if (null? prefix) '(command) prefix) (list name)))
   (define (prev-prefix prefix)
-    (cond ((and (= 2 (length prefix))))
+    (cond ((and (= 2 (length prefix))) '())
           ((null? prefix) '())
           (else (reverse (cdr (reverse  prefix))))))
   (let ((fail (if (pair? o)
@@ -346,13 +347,24 @@
     (cond
      ((null? spec)
       (error "no procedure in application spec"))
+     ((or (null? (car spec)) (equal? '(@) (car spec)))
+      (parse-app prefix (cdr spec) opt-spec args config init end fail))
      ((pair? (car spec))
       (case (caar spec)
         ((@)
-         (let* ((new-opt-spec (cadr (car spec)))
+         (let* ((tail (cdar spec))
+                (new-opt-spec
+                 (cond
+                  ((not (pair? tail))
+                   '())
+                  ((or (pair? (cdr tail))
+                       (and (pair? (car tail)) (symbol? (caar tail))))
+                   tail)
+                  (else
+                   (car tail))))
                 (new-fail
                  (lambda (new-prefix new-spec new-opt new-args reason)
-                   (parse-options (prev-prefix prefix) opt-spec new-args fail)))
+                   (parse-option (prev-prefix prefix) opt-spec new-args fail)))
                 (cfg+args (parse-options prefix new-opt-spec args new-fail))
                 (config (conf-append (car cfg+args) config))
                 (args (cdr cfg+args)))
