@@ -8,11 +8,12 @@
 (define (cddr x) (cdr (cdr x)))
 
 (define (cons-source kar kdr source)
-  ((lambda (pair)
+  ((lambda (pair source)
      (if (pair? source)
          (pair-source-set! pair (pair-source source)))
      pair)
-   (cons kar kdr)))
+   (cons kar kdr)
+   (strip-syntactic-closures source)))
 
 ;; basic utils
 
@@ -194,7 +195,8 @@
                (cadr (car x))
                (list (rename 'append) (cadr (car x)) (qq (cdr x) d))))
           (else
-           (list (rename 'cons) (qq (car x) d) (qq (cdr x) d)))))
+           (list (rename 'cons-source) (qq (car x) d) (qq (cdr x) d)
+                 (list (rename 'quote) x)))))
         ((vector? x) (list (rename 'list->vector) (qq (vector->list x) d)))
         ((if (identifier? x) #t (null? x)) (list (rename 'quote) x))
         (else x)))
@@ -394,12 +396,27 @@
    ((if (null? o) #t (eq? 10 (car o)))
     (%number->string num))
    (else
-    (let lp ((n (abs num)) (d (car o)) (res '()))
-      (if (> n 0)
-          (lp (quotient n d) d (cons (digit-char (remainder n d)) res))
-          (if (null? res)
-              "0"
-              (list->string (if (negative? num) (cons #\- res) res))))))))
+    (let ((d (car o)))
+      (cond
+       ((%complex? num)
+        (let ((real (real-part num))
+              (imag (imag-part num)))
+          (string-append (number->string real d) (if (negative? imag) "-" "+")
+                         (number->string imag d) "i")))
+       ((inexact? num)
+        (string-append "#i" (number->string (inexact->exact num) d)))
+       ((ratio? num)
+        (string-append (number->string (numerator num) d) "/"
+                       (number->string (denominator num) d)))
+       (else
+        (let lp ((n (abs num)) (res '()))
+          (cond
+           ((> n 0)
+            (lp (quotient n d) (cons (digit-char (remainder n d)) res)))
+           ((null? res)
+            "0")
+           (else
+            (list->string (if (negative? num) (cons #\- res) res)))))))))))
 
 (define (list->string ls)
   (call-with-output-string
@@ -408,10 +425,11 @@
 (define (string->list str . o)
   (cond
    ((null? o)
-    (let lp ((i (string-cursor-prev str (string-cursor-end str))) (res '()))
-      (if (< i 0)
-          res
-          (lp (string-cursor-prev str i) (cons (string-cursor-ref str i) res)))))
+    (let ((start (string-cursor-start str)))
+      (let lp ((i (string-cursor-prev str (string-cursor-end str))) (res '()))
+        (if (string-cursor<? i start)
+            res
+            (lp (string-cursor-prev str i) (cons (string-cursor-ref str i) res))))))
    (else
     (string->list (apply substring str o)))))
 
@@ -669,231 +687,250 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; syntax-rules
 
+(define (syntax-rules-transformer expr rename compare)
+  (let ((ellipsis-specified? (identifier? (cadr expr)))
+        (count 0)
+        (_er-macro-transformer (rename 'er-macro-transformer))
+        (_lambda (rename 'lambda))      (_let (rename 'let))
+        (_begin (rename 'begin))        (_if (rename 'if))
+        (_and (rename 'and))            (_or (rename 'or))
+        (_eq? (rename 'eq?))            (_equal? (rename 'equal?))
+        (_car (rename 'car))            (_cdr (rename 'cdr))
+        (_cons (rename 'cons))          (_pair? (rename 'pair?))
+        (_null? (rename 'null?))        (_expr (rename 'expr))
+        (_rename (rename 'rename))      (_compare (rename 'compare))
+        (_quote (rename 'syntax-quote)) (_apply (rename 'apply))
+        (_append (rename 'append))      (_map (rename 'map))
+        (_vector? (rename 'vector?))    (_list? (rename 'list?))
+        (_len (rename'len))             (_length (rename 'length*))
+        (_- (rename '-))   (_>= (rename '>=))   (_error (rename 'error))
+        (_ls (rename 'ls)) (_res (rename 'res)) (_i (rename 'i))
+        (_reverse (rename 'reverse))
+        (_vector->list (rename 'vector->list))
+        (_list->vector (rename 'list->vector))
+        (_cons3 (rename 'cons-source))
+        (_underscore (rename '_)))
+    (define ellipsis (if ellipsis-specified? (cadr expr) (rename '...)))
+    (define lits (if ellipsis-specified? (car (cddr expr)) (cadr expr)))
+    (define forms (if ellipsis-specified? (cdr (cddr expr)) (cddr expr)))
+    (define (next-symbol s)
+      (set! count (+ count 1))
+      (rename (string->symbol (string-append s (%number->string count)))))
+    (define (expand-pattern pat tmpl)
+      (let lp ((p (cdr pat))
+               (x (list _cdr _expr))
+               (dim 0)
+               (vars '())
+               (k (lambda (vars)
+                    (list _cons (expand-template tmpl vars) #f))))
+        (let ((v (next-symbol "v.")))
+          (list
+           _let (list (list v x))
+           (cond
+            ((identifier? p)
+             (cond
+              ((ellipsis-mark? p)
+               (error "bad ellipsis" p))
+              ((memq p lits)
+               (list _and
+                     (list _compare v (list _rename (list _quote p)))
+                     (k vars)))
+              ((compare p _underscore)
+               (k vars))
+              (else
+               (list _let (list (list p v)) (k (cons (cons p dim) vars))))))
+            ((ellipsis? p)
+             (cond
+              ((not (null? (cdr (cdr p))))
+               (cond
+                ((any (lambda (x) (and (identifier? x) (ellipsis-mark? x)))
+                      (cddr p))
+                 (error "multiple ellipses" p))
+                (else
+                 (let ((len (length* (cdr (cdr p))))
+                       (_lp (next-symbol "lp.")))
+                   `(,_let ((,_len (,_length ,v)))
+                      (,_and (,_>= ,_len ,len)
+                             (,_let ,_lp ((,_ls ,v)
+                                          (,_i (,_- ,_len ,len))
+                                          (,_res (,_quote ())))
+                                    (,_if (,_>= 0 ,_i)
+                                        ,(lp `(,(cddr p)
+                                               (,(car p) ,(car (cdr p))))
+                                             `(,_cons ,_ls
+                                                      (,_cons (,_reverse ,_res)
+                                                              (,_quote ())))
+                                             dim
+                                             vars
+                                             k)
+                                        (,_lp (,_cdr ,_ls)
+                                              (,_- ,_i 1)
+                                              (,_cons3 (,_car ,_ls)
+                                                       ,_res
+                                                       ,_ls))))))))))
+              ((identifier? (car p))
+               (list _and (list _list? v)
+                     (list _let (list (list (car p) v))
+                           (k (cons (cons (car p) (+ 1 dim)) vars)))))
+              (else
+               (let* ((w (next-symbol "w."))
+                      (_lp (next-symbol "lp."))
+                      (new-vars (all-vars (car p) (+ dim 1)))
+                      (ls-vars (map (lambda (x)
+                                      (next-symbol
+                                       (string-append
+                                        (symbol->string
+                                         (identifier->symbol (car x)))
+                                        "-ls")))
+                                    new-vars))
+                      (once
+                       (lp (car p) (list _car w) (+ dim 1) '()
+                           (lambda (_)
+                             (cons
+                              _lp
+                              (cons
+                               (list _cdr w)
+                               (map (lambda (x l)
+                                      (list _cons (car x) l))
+                                    new-vars
+                                    ls-vars)))))))
+                 (list
+                  _let
+                  _lp (cons (list w v)
+                            (map (lambda (x) (list x (list _quote '()))) ls-vars))
+                  (list _if (list _null? w)
+                        (list _let (map (lambda (x l)
+                                          (list (car x) (list _reverse l)))
+                                        new-vars
+                                        ls-vars)
+                              (k (append new-vars vars)))
+                        (list _and (list _pair? w) once)))))))
+            ((pair? p)
+             (list _and (list _pair? v)
+                   (lp (car p)
+                       (list _car v)
+                       dim
+                       vars
+                       (lambda (vars)
+                         (lp (cdr p) (list _cdr v) dim vars k)))))
+            ((vector? p)
+             (list _and
+                   (list _vector? v)
+                   (lp (vector->list p) (list _vector->list v) dim vars k)))
+            ((null? p) (list _and (list _null? v) (k vars)))
+            (else (list _and (list _equal? v p) (k vars))))))))
+    (define ellipsis-mark?
+      (if (if ellipsis-specified?
+              (memq ellipsis lits)
+              (any (lambda (x) (compare ellipsis x)) lits))
+          (lambda (x) #f)
+          (if ellipsis-specified?
+              (lambda (x) (eq? ellipsis x))
+              (lambda (x) (compare ellipsis x)))))
+    (define (ellipsis-escape? x) (and (pair? x) (ellipsis-mark? (car x))))
+    (define (ellipsis? x)
+      (and (pair? x) (pair? (cdr x)) (ellipsis-mark? (cadr x))))
+    (define (ellipsis-depth x)
+      (if (ellipsis? x)
+          (+ 1 (ellipsis-depth (cdr x)))
+          0))
+    (define (ellipsis-tail x)
+      (if (ellipsis? x)
+          (ellipsis-tail (cdr x))
+          (cdr x)))
+    (define (all-vars x dim)
+      (let lp ((x x) (dim dim) (vars '()))
+        (cond ((identifier? x)
+               (if (or (memq x lits)
+                       (compare x _underscore))
+                   vars
+                   (cons (cons x dim) vars)))
+              ((ellipsis? x) (lp (car x) (+ dim 1) (lp (cddr x) dim vars)))
+              ((pair? x) (lp (car x) dim (lp (cdr x) dim vars)))
+              ((vector? x) (lp (vector->list x) dim vars))
+              (else vars))))
+    (define (free-vars x vars dim)
+      (let lp ((x x) (free '()))
+        (cond
+         ((identifier? x)
+          (if (and (not (memq x free))
+                   (cond ((assq x vars) => (lambda (cell) (>= (cdr cell) dim)))
+                         (else #f)))
+              (cons x free)
+              free))
+         ((pair? x) (lp (car x) (lp (cdr x) free)))
+         ((vector? x) (lp (vector->list x) free))
+         (else free))))
+    (define (expand-template tmpl vars)
+      (let lp ((t tmpl) (dim 0))
+        (cond
+         ((identifier? t)
+          (cond
+           ((find (lambda (v) (eq? t (car v))) vars)
+            => (lambda (cell)
+                 (if (<= (cdr cell) dim)
+                     t
+                     (error "too few ...'s"))))
+           (else
+            (list _rename (list _quote t)))))
+         ((pair? t)
+          (cond
+           ((ellipsis-escape? t)
+            (list _quote
+                  (if (pair? (cdr t))
+                      (if (pair? (cddr t)) (cddr t) (cadr t))
+                      (cdr t))))
+           ((ellipsis? t)
+            (let* ((depth (ellipsis-depth t))
+                   (ell-dim (+ dim depth))
+                   (ell-vars (free-vars (car t) vars ell-dim)))
+              (cond
+               ((null? ell-vars)
+                (error "too many ...'s"))
+               ((and (null? (cdr (cdr t))) (identifier? (car t)))
+                ;; shortcut for (var ...)
+                (lp (car t) ell-dim))
+               (else
+                (let* ((once (lp (car t) ell-dim))
+                       (nest (if (and (null? (cdr ell-vars))
+                                      (identifier? once)
+                                      (eq? once (car vars)))
+                                 once ;; shortcut
+                                 (cons _map
+                                       (cons (list _lambda ell-vars once)
+                                             ell-vars))))
+                       (many (do ((d depth (- d 1))
+                                  (many nest
+                                        (list _apply _append many)))
+                                 ((= d 1) many))))
+                  (if (null? (ellipsis-tail t))
+                      many ;; shortcut
+                      (list _append many (lp (ellipsis-tail t) dim))))))))
+           (else (list _cons3 (lp (car t) dim) (lp (cdr t) dim) (list _quote t)))))
+         ((vector? t) (list _list->vector (lp (vector->list t) dim)))
+         ((null? t) (list _quote '()))
+         (else t))))
+    (list
+     _er-macro-transformer
+     (list _lambda (list _expr _rename _compare)
+           (list
+            _car
+            (cons
+             _or
+             (append
+              (map
+               (lambda (clause) (expand-pattern (car clause) (cadr clause)))
+               forms)
+              (list
+               (list _cons
+                     (list _error "no expansion for"
+                           (list (rename 'strip-syntactic-closures) _expr))
+                     #f)))))))))
+
 (define-syntax syntax-rules
   (er-macro-transformer
    (lambda (expr rename compare)
-     (let ((ellipsis-specified? (identifier? (cadr expr)))
-           (count 0)
-           (_er-macro-transformer (rename 'er-macro-transformer))
-           (_lambda (rename 'lambda))      (_let (rename 'let))
-           (_begin (rename 'begin))        (_if (rename 'if))
-           (_and (rename 'and))            (_or (rename 'or))
-           (_eq? (rename 'eq?))            (_equal? (rename 'equal?))
-           (_car (rename 'car))            (_cdr (rename 'cdr))
-           (_cons (rename 'cons))          (_pair? (rename 'pair?))
-           (_null? (rename 'null?))        (_expr (rename 'expr))
-           (_rename (rename 'rename))      (_compare (rename 'compare))
-           (_quote (rename 'syntax-quote)) (_apply (rename 'apply))
-           (_append (rename 'append))      (_map (rename 'map))
-           (_vector? (rename 'vector?))    (_list? (rename 'list?))
-           (_len (rename'len))             (_length (rename 'length*))
-           (_- (rename '-))   (_>= (rename '>=))   (_error (rename 'error))
-           (_ls (rename 'ls)) (_res (rename 'res)) (_i (rename 'i))
-           (_reverse (rename 'reverse))
-           (_vector->list (rename 'vector->list))
-           (_list->vector (rename 'list->vector))
-           (_cons3 (rename 'cons-source)))
-       (define ellipsis (rename (if ellipsis-specified? (cadr expr) '...)))
-       (define lits (if ellipsis-specified? (car (cddr expr)) (cadr expr)))
-       (define forms (if ellipsis-specified? (cdr (cddr expr)) (cddr expr)))
-       (define (next-symbol s)
-         (set! count (+ count 1))
-         (rename (string->symbol (string-append s (%number->string count)))))
-       (define (expand-pattern pat tmpl)
-         (let lp ((p (cdr pat))
-                  (x (list _cdr _expr))
-                  (dim 0)
-                  (vars '())
-                  (k (lambda (vars)
-                       (list _cons (expand-template tmpl vars) #f))))
-           (let ((v (next-symbol "v.")))
-             (list
-              _let (list (list v x))
-              (cond
-               ((identifier? p)
-                (if (any (lambda (l) (compare p l)) lits)
-                    (list _and
-                          (list _compare v (list _rename (list _quote p)))
-                          (k vars))
-                    (list _let (list (list p v)) (k (cons (cons p dim) vars)))))
-               ((ellipsis? p)
-                (cond
-                 ((not (null? (cdr (cdr p))))
-                  (cond
-                   ((any (lambda (x) (and (identifier? x) (compare x ellipsis)))
-                         (cddr p))
-                    (error "multiple ellipses" p))
-                   (else
-                    (let ((len (length* (cdr (cdr p))))
-                          (_lp (next-symbol "lp.")))
-                      `(,_let ((,_len (,_length ,v)))
-                         (,_and (,_>= ,_len ,len)
-                                (,_let ,_lp ((,_ls ,v)
-                                             (,_i (,_- ,_len ,len))
-                                             (,_res (,_quote ())))
-                                  (,_if (,_>= 0 ,_i)
-                                      ,(lp `(,(cddr p) 
-                                             (,(car p) ,(car (cdr p))))
-                                           `(,_cons ,_ls
-                                                    (,_cons (,_reverse ,_res)
-                                                            (,_quote ())))
-                                           dim
-                                           vars
-                                           k)
-                                      (,_lp (,_cdr ,_ls)
-                                            (,_- ,_i 1)
-                                            (,_cons3 (,_car ,_ls)
-                                                     ,_res
-                                                     ,_ls))))))))))
-                 ((identifier? (car p))
-                  (list _and (list _list? v)
-                        (list _let (list (list (car p) v))
-                              (k (cons (cons (car p) (+ 1 dim)) vars)))))
-                 (else
-                  (let* ((w (next-symbol "w."))
-                         (_lp (next-symbol "lp."))
-                         (new-vars (all-vars (car p) (+ dim 1)))
-                         (ls-vars (map (lambda (x)
-                                         (next-symbol
-                                          (string-append
-                                           (symbol->string
-                                            (identifier->symbol (car x)))
-                                           "-ls")))
-                                       new-vars))
-                         (once
-                          (lp (car p) (list _car w) (+ dim 1) '()
-                              (lambda (_)
-                                (cons
-                                 _lp
-                                 (cons
-                                  (list _cdr w)
-                                  (map (lambda (x l)
-                                         (list _cons (car x) l))
-                                       new-vars
-                                       ls-vars)))))))
-                    (list
-                     _let
-                     _lp (cons (list w v)
-                               (map (lambda (x) (list x (list _quote '()))) ls-vars))
-                     (list _if (list _null? w)
-                           (list _let (map (lambda (x l)
-                                             (list (car x) (list _reverse l)))
-                                           new-vars
-                                           ls-vars)
-                                 (k (append new-vars vars)))
-                           (list _and (list _pair? w) once)))))))
-               ((pair? p)
-                (list _and (list _pair? v)
-                      (lp (car p)
-                          (list _car v)
-                          dim
-                          vars
-                          (lambda (vars)
-                            (lp (cdr p) (list _cdr v) dim vars k)))))
-               ((vector? p)
-                (list _and
-                      (list _vector? v)
-                      (lp (vector->list p) (list _vector->list v) dim vars k)))
-               ((null? p) (list _and (list _null? v) (k vars)))
-               (else (list _and (list _equal? v p) (k vars))))))))
-       (define (ellipsis-escape? x) (and (pair? x) (compare ellipsis (car x))))
-       (define (ellipsis? x)
-         (and (pair? x) (pair? (cdr x)) (compare ellipsis (cadr x))))
-       (define (ellipsis-depth x)
-         (if (ellipsis? x)
-             (+ 1 (ellipsis-depth (cdr x)))
-             0))
-       (define (ellipsis-tail x)
-         (if (ellipsis? x)
-             (ellipsis-tail (cdr x))
-             (cdr x)))
-       (define (all-vars x dim)
-         (let lp ((x x) (dim dim) (vars '()))
-           (cond ((identifier? x)
-                  (if (any (lambda (lit) (compare x lit)) lits)
-                      vars
-                      (cons (cons x dim) vars)))
-                 ((ellipsis? x) (lp (car x) (+ dim 1) (lp (cddr x) dim vars)))
-                 ((pair? x) (lp (car x) dim (lp (cdr x) dim vars)))
-                 ((vector? x) (lp (vector->list x) dim vars))
-                 (else vars))))
-       (define (free-vars x vars dim)
-         (let lp ((x x) (free '()))
-           (cond
-            ((identifier? x)
-             (if (and (not (memq x free))
-                      (cond ((assq x vars) => (lambda (cell) (>= (cdr cell) dim)))
-                            (else #f)))
-                 (cons x free)
-                 free))
-            ((pair? x) (lp (car x) (lp (cdr x) free)))
-            ((vector? x) (lp (vector->list x) free))
-            (else free))))
-       (define (expand-template tmpl vars)
-         (let lp ((t tmpl) (dim 0))
-           (cond
-            ((identifier? t)
-             (cond
-              ((find (lambda (v) (compare t (car v))) vars)
-               => (lambda (cell)
-                    (if (<= (cdr cell) dim)
-                        t
-                        (error "too few ...'s"))))
-              (else
-               (list _rename (list _quote t)))))
-            ((pair? t)
-             (cond
-              ((ellipsis-escape? t)
-               (list _quote
-                     (if (pair? (cdr t))
-                         (if (pair? (cddr t)) (cddr t) (cadr t))
-                         (cdr t))))
-              ((ellipsis? t)
-               (let* ((depth (ellipsis-depth t))
-                      (ell-dim (+ dim depth))
-                      (ell-vars (free-vars (car t) vars ell-dim)))
-                 (cond
-                  ((null? ell-vars)
-                   (error "too many ...'s"))
-                  ((and (null? (cdr (cdr t))) (identifier? (car t)))
-                   ;; shortcut for (var ...)
-                   (lp (car t) ell-dim))
-                  (else
-                   (let* ((once (lp (car t) ell-dim))
-                          (nest (if (and (null? (cdr ell-vars))
-                                         (identifier? once)
-                                         (eq? once (car vars)))
-                                    once ;; shortcut
-                                    (cons _map
-                                          (cons (list _lambda ell-vars once)
-                                                ell-vars))))
-                          (many (do ((d depth (- d 1))
-                                     (many nest
-                                           (list _apply _append many)))
-                                    ((= d 1) many))))
-                     (if (null? (ellipsis-tail t))
-                         many ;; shortcut
-                         (list _append many (lp (ellipsis-tail t) dim))))))))
-              (else (list _cons3 (lp (car t) dim) (lp (cdr t) dim) (list _quote t)))))
-            ((vector? t) (list _list->vector (lp (vector->list t) dim)))
-            ((null? t) (list _quote '()))
-            (else t))))
-       (list
-        _er-macro-transformer
-        (list _lambda (list _expr _rename _compare)
-              (list
-               _car
-               (cons
-                _or
-                (append
-                 (map
-                  (lambda (clause) (expand-pattern (car clause) (cadr clause)))
-                  forms)
-                 (list
-                  (list _cons
-                        (list _error "no expansion for"
-                              (list (rename 'strip-syntactic-closures) _expr))
-                        #f)))))))))))
+     (syntax-rules-transformer expr rename compare))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; additional syntax
@@ -1048,8 +1085,11 @@
  (auto-force
   )
  (else
+  (define *promise-tag* (list 'promise))
   (define (promise done? proc)
-    (list (cons done? proc)))
+    (cons (cons done? proc) *promise-tag*))
+  (define (promise? x)
+    (and (pair? x) (eq? *promise-tag* (cdr x))))
   (define (promise-done? x) (car (car x)))
   (define (promise-value x) (cdr (car x)))
   (define (promise-update! new old)
@@ -1072,6 +1112,7 @@
   (define (exact-complex? x)
     (and (%complex? x) (exact? (complex-real x)) (exact? (complex-imag x)))))
  (else
+  (define (%complex? x) #f)
   (define (exact-complex? x) #f)))
 
 (cond-expand
@@ -1097,6 +1138,7 @@
             1.0
             (exact->inexact (denominator (inexact->exact x)))))))
  (else
+  (define (ratio? x) #f)
   (cond-expand
    (complex
     (define (exact? x)
@@ -1226,25 +1268,30 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; string cursors
 
-(define string-cursor<? <)
-(define string-cursor<=? <=)
-(define string-cursor>? >)
-(define string-cursor>=? >=)
-(define string-cursor=? =)
-
-(define (string-cursor-start s) 0)
-
 (define (string-copy str . o)
   (apply substring str (if (pair? o) o '(0))))
 
+(define string-cursor=? eq?)
+
 (cond-expand
  (full-unicode
-  (define string-cursor-end string-size))
+  (define string-cursor-start
+    (let ((start (string-index->cursor "" 0)))
+      (lambda (s) start)))
+  (define (string-size s)
+    (string-cursor-offset (string-cursor-end s))))
  (else
-  (define (string-index->offset str i) i)
-  (define (string-offset->index str off) off)
+  (define string-cursor? fixnum?)
+  (define string-cursor<? <)
+  (define string-cursor<=? <=)
+  (define string-cursor>? >)
+  (define string-cursor>=? >=)
+  (define (string-index->cursor str i) i)
+  (define (string-cursor->index str off) off)
+  (define (string-cursor-offset str off) off)
   (define string-size string-length)
   (define substring-cursor substring)
+  (define (string-cursor-start s) 0)
   (define string-cursor-end string-length)
   (define string-cursor-ref string-ref)
   (define (string-cursor-next s i) (+ i 1))
